import numpy as np

def generateNeighboursInBall(vec: list, nb_neighbours, ball):
    ball_low, ball_high = ball
    neighbours = [(np.array(vec) * (1 + np.random.uniform(ball_low, ball_high)/100)).tolist() for _ in range(nb_neighbours)]

    if nb_neighbours != 1:
        return neighbours
    else:
        return neighbours[0]

def isInBall(vec: list, center: list, ball):
    """Check if the vector is inside a ball centered at another vector"""
    ball_low, ball_high = center * (1 + ball/100)
    if np.all(vec >= ball_low) and np.all(vec <= ball_high): 
        return True
    return False


import itertools
import numpy as np
from scipy.spatial import ConvexHull


def generateUniformDirectionsOnCube(nbrStepsForEachDirection):
    """Creates points placed uniformly on the cube"""
    all_xs = np.linspace(-1, 1, nbrStepsForEachDirection)
    all_ys = np.linspace(-1, 1, nbrStepsForEachDirection)
    all_zs = np.linspace(-1, 1, nbrStepsForEachDirection)

    directions = [all_xs, all_ys, all_zs]
    all_directions = list(itertools.product(*directions))
    if (0, 0, 0) in all_directions:
        all_directions.remove((0, 0, 0))
    all_directions = np.array(all_directions)

    return all_directions


def intersectConvexHullWithDirection(hull, direction):
    """ The sense of the vector direction matters! """
    eq = hull.equations.T
    V, b = eq[:-1], eq[-1]
    alpha = -b/np.dot(V.T, direction)
    return np.min(alpha[alpha > 0]) * direction

def intersectPolytopeWithDirections(polytopeVertices, directions):
    """Give the intersection points in the order of the given directions"""
    hull = ConvexHull(points=polytopeVertices.T)

    pts = []
    for direction in directions:
        pt = intersectConvexHullWithDirection(hull, direction)
        pts.append(pt)
    pts = np.array(pts)

    return pts

def discretizeIn26Vertices(vertices):
    try:
        directions = generateUniformDirectionsOnCube(3)
        intersections = intersectPolytopeWithDirections(vertices, directions)
    except:
        intersections = np.zeros((26, 3))

    return intersections


def getDiscretizedNormVector26(vertices):
    """Discretize a polytope in 26 uniform directions then retrieve the norm of each new vertice"""
    vertices_disc = discretizeIn26Vertices(vertices)
    return np.linalg.norm(vertices_disc, axis=1)


def getEigenValuesVectors(A: np.ndarray):
    """From a list of vertices, returns the (normalized) axes of the ellipsoid (eigenvectors) and the eigen values associated
    obtained from covariance matrix.

    Args:
        A (np.ndarray): list of vertices (columns or rows)

    Returns:
        _type_: eigenvalues in decreasing order + associated eigenvectors as columns of v
    """

    cov = A@A.T
    v, w, _ = np.linalg.svd(cov)
    eigen_values = np.sqrt(w)
    eigen_vectors = v

    return eigen_values, eigen_vectors


import numpy as np
import biorbd


def getModelAsVector(model: biorbd.Model) -> list:
    vec = []
    for m in range(model.nbMuscles()):
        muscle = model.muscle(m)
        c = muscle.characteristics()
        vec.append(c.forceIsoMax())
        vec.append(c.optimalLength())
        vec.append(c.tendonSlackLength())

        g = muscle.position()
        o = g.originInLocal().to_array().tolist()
        i = g.insertionInLocal().to_array().tolist()
        vec += o + i
    return vec


def getModelFromVector(vec: list, model_path: str) -> biorbd.Model:
    vec = np.array(vec)
    model = biorbd.Model(model_path)
    for m in range(model.nbMuscles()):
        muscle = model.muscle(m)

        c = muscle.characteristics()
        c.setForceIsoMax(vec[m*9+0])
        c.setOptimalLength(vec[m*9+1])
        c.setTendonSlackLength(vec[m*9+2])

        g = muscle.position()
        g.setOrigin(np.array(vec[m*9+3: m*9+6]))
        g.setInsertionInLocal(np.array(vec[m*9+6: m*9+9]))
    return model


def getMomentArmMatrix(model: biorbd.Model, Q: np.ndarray) -> np.ndarray:
    """Return L, not -L.T"""
    return model.musclesLengthJacobian(Q).to_array()


def getStationJacobian(model, Q, bodyName: str, endEffector: list):
    x, y, z = endEffector
    point = biorbd.NodeSegment(biorbd.Vector3d(x, y, z))
    return model.markersJacobian(Q, biorbd.String(bodyName), point, False).to_array()


def getMuscleTensions(model):
    nbMuscles = model.nbMuscles()

    # ligne 1 = tensions min, ligne 2 = tensions max
    tensions = np.zeros((2, nbMuscles))
    for i in range(nbMuscles):
        t_min, t_max = getMuscleTensionsId(model, i)
        tensions[0, i] = t_min
        tensions[1, i] = t_max

    np.nan_to_num(tensions, copy=False)

    return tensions[0, :], tensions[1, :]


def getMuscleTensionsId(model, muscleId: int):
    muscle = biorbd.HillThelenType(model.muscle(muscleId))
    muscleState = muscle.state()

    a_ce = muscle.FlCE(muscleState)
    a_pe = muscle.FlPE()

    if a_ce > 1:
        a_ce = 1
    elif a_ce < 0:
        a_ce = 0

    if a_pe > 1:
        a_pe = 0.1
    elif a_pe < 0:
        a_pe = 0

    forceIsoMax = muscle.characteristics().forceIsoMax()
    if forceIsoMax < 0:
        forceIsoMax = 0
    t_min = a_pe * forceIsoMax
    t_max = (a_ce + a_pe) * forceIsoMax

    return t_min, t_max


import numpy as np
import biorbd
import pycapacity.human as capacity
import identification_study.utils.model_computation as mcomp
from pycapacity.algorithms import iterative_convex_hull_method

def getPolytope(model: biorbd.Model, Q: np.ndarray, bodyName: str, endEffector: list, tol=1, withBias: bool = True, name: str = None) -> dict:
    zeros = np.zeros(model.nbQ())

    model.updateMuscles(Q, True)
    model.UpdateKinematicsCustom(Q, zeros, zeros)

    t_min, t_max = mcomp.getMuscleTensions(model)
    N = -mcomp.getMomentArmMatrix(model, Q).T
    J = mcomp.getStationJacobian(model, Q, bodyName, endEffector)

    error = False
    vertices = np.array([])
    face_indices = np.array([])
    muscle_force_vertices = np.array([])
    try:
        if withBias:
            Tau_grav = model.InverseDynamics(Q, zeros, zeros)
            pol = capacity.force_polytope(
                J, N, t_min, t_max, tol, -Tau_grav.to_array(), options = { "calculate_faces": True })
            vertices = pol.vertices
            face_indices = pol.face_indices
            muscle_force_vertices = pol.mucsle_force_vertices
        else:
            pol = capacity.force_polytope(
                J, N, t_min, t_max, tol, options = { "calculate_faces": True })
            vertices = pol.vertices
            face_indices = pol.face_indices
    except Exception as e:
        # print("Error in polytope computation.")
        # print(e.args)
        error = True

    toReturn = {
        "name": name,
        "error": error,
        "J": J,
        "N": N,
        "t_min": t_min,
        "t_max": t_max,
        "tol": tol,
        "vertices": vertices,
        "face_indices": face_indices,
        "muscle_force_vertices": muscle_force_vertices
    }
    return toReturn


def getPolytopeInDirection(model: biorbd.Model, Q: np.ndarray, bodyName: str, endEffector: list, direction: np.ndarray, tol=1, withBias: bool = True, name: str = None) -> dict:
    zeros = np.zeros(model.nbQ())

    model.updateMuscles(Q, True)
    model.UpdateKinematicsCustom(Q, zeros, zeros)

    t_min, t_max = mcomp.getMuscleTensions(model)
    N = -mcomp.getMomentArmMatrix(model, Q).T
    J = mcomp.getStationJacobian(model, Q, bodyName, endEffector)

    error = False
    vertices = np.array([])
    face_indices = np.array([])
    try:
        if withBias:
            Tau_grav = model.InverseDynamics(Q, zeros, zeros)
            vertices, H, d, face_indices, _, _ = iterative_convex_hull_method(direction@J.T, N, t_min, t_max, tol, -Tau_grav.to_array())
            # vertices, H, d, face_indices = capacity.force_polytope(
            #     direction@J, N, t_min, t_max, tol, -Tau_grav.to_array())
        else:
            # vertices, H, d, face_indices = capacity.force_polytope(
            #     J, N, t_min, t_max, tol)
            vertices, H, d, face_indices, _, _ = iterative_convex_hull_method(direction@J.T, N, t_min, t_max, tol)
    except Exception as e:
        # print("Error in polytope computation.")
        # print(e.args)
        error = True

    toReturn = {
        "name": name,
        "error": error,
        "J": J,
        "N": N,
        "t_min": t_min,
        "t_max": t_max,
        "tol": tol,
        "vertices": vertices,
        "face_indices": face_indices
    }
    return toReturn


import numpy as np
import matplotlib.pyplot as plt

# 1. Directions spanned by the octahedron vertices (3)
def createD1(a1 = 1, a2 = 1, a3 = 1):
    """a1, a2, a3 are respectively the lengths of e1, e2, e3"""
    main_dir = np.diag([a1,a2,a3])
    all_dir = np.vstack([main_dir, -main_dir])
    return main_dir, all_dir

def createD2(a1 = 1, a2 = 1, a3 = 1):
    main_oct, _ = createD1(a1,a2,a3)

    e1, e2, e3 = main_oct
    pt12 = e1 + e2
    pt23 = e2 + e3
    pt13 = e1 + e3
    bar1 = (pt12 + pt23 + pt13)/3
    bar2 = np.hstack([bar1[0], bar1[1], -bar1[2]])
    bar3 = np.hstack([bar1[0], -bar1[1], bar1[2]])
    bar4 = np.hstack([bar1[0], -bar1[1], -bar1[2]])
    bar5 = np.hstack([-bar1[0], bar1[1], bar1[2]])
    bar6 = np.hstack([-bar1[0], bar1[1], -bar1[2]])
    bar7 = np.hstack([-bar1[0], -bar1[1], bar1[2]])
    bar8 = np.hstack([-bar1[0], -bar1[1], -bar1[2]])

    all_dir = np.vstack([bar1, bar2, bar3, bar4, bar5, bar6, bar7, bar8])
    return all_dir

def createD3(a1 = 1, a2 = 1, a3 = 1):
    main_oct, _ = createD1(a1,a2,a3)
    e1, e2, e3 = main_oct
    
    pt1_e12 =  e1 + e2
    pt2_e12 =  e1 - e2
    pt3_e12 = -e1 + e2
    pt4_e12 = -e1 - e2

    pt1_e23 =  e2 + e3
    pt2_e23 =  e2 - e3
    pt3_e23 = -e2 + e3
    pt4_e23 = -e2 - e3

    pt1_e13 =  e1 + e3
    pt2_e13 =  e1 - e3
    pt3_e13 = -e1 + e3
    pt4_e13 = -e1 - e3

    all_dir = np.vstack([
        pt1_e12,
        pt2_e12,
        pt3_e12,
        pt4_e12,
        pt1_e23,
        pt2_e23,
        pt3_e23,
        pt4_e23,
        pt1_e13,
        pt2_e13,
        pt3_e13,
        pt4_e13
    ])

    return all_dir
D3 = createD3(1,2,3)

def getD1(a1=1, a2=1, a3=1):
    _, D1 = createD1(a1, a2, a3)
    return D1

def getD2(a1=1, a2=1, a3=1):
    D2 = createD2(a1, a2, a3)
    return D2

def getD3(a1=1, a2=1, a3=1):
    D3 = createD3(a1, a2, a3)
    return D3

# Directions are cube vertices + octahedron vertices
# 7 dir = 14 vertices
def getD4(a1=1, a2=1, a3=1):
    D1 = getD1(a1, a2, a3)
    D2 = getD2(a1, a2, a3)
    return np.vstack([D1, D2])

# Directions are octahedron vertices + rectangle planes vertices
# 9 dir = 18 vertices
def getD5(a1=1, a2=1, a3=1):
    D1 = getD1(a1, a2, a3)
    D3 = getD3(a1, a2, a3)
    return np.vstack([D1, D3])

# Directions are cube vertices + rectangle planes vertices
# 10 dir = 20 vertices
def getD6(a1=1, a2=1, a3=1):
    D2 = getD2(a1, a2, a3)
    D3 = getD3(a1, a2, a3)
    return np.vstack([D2, D3])

# Directions are cube vertices + octahedron vertices + rectangle planes vertices
# 13 dir = 26 vertices
def getD7(a1=1, a2=1, a3=1):
    D1 = getD1(a1, a2, a3)
    D2 = getD2(a1, a2, a3)
    D3 = getD3(a1, a2, a3)
    return np.vstack([D1, D2, D3])




import numpy as np
import matplotlib.pyplot as plt
import itertools

def split_sphere(R = 1, horizontal_split = 36, vertical_split = 36, method="equal_angles"):
    theta = np.linspace(0,360,horizontal_split+1)
    if method == "equal_angles":
        phi = np.linspace(0, 180, vertical_split+1)
        c = np.cos(phi)
        s = np.sin(phi)
    elif method == "equal_area":
        c = np.linspace(-1, 1, vertical_split+1)
        s = 1 - c**2
    else:
        raise(ValueError('method must be "equal_angles" or "equal_area"'))
    x = R * np.outer(s, np.cos(theta))
    y = R * np.outer(s, np.sin(theta))
    z = R * np.outer(c, np.ones(horizontal_split+1))
    return x, y, z

def sphere_parametric(nb_theta = 36, nb_phi = 36, r = 1, center = np.zeros(3)):
    thetas = np.linspace(0, np.pi, nb_theta)
    phis = np.linspace(0, 2*np.pi, nb_phi)

    comb = itertools.product(thetas, phis)

    x0, y0, z0 = center

    points = []
    for theta, phi in comb:
        x = x0 + r*np.sin(theta)*np.cos(phi)
        y = y0 + r*np.sin(theta)*np.sin(phi)
        z = z0 + r*np.cos(theta)
        point = np.array([x,y,z])
        points.append(point)
    
    return np.array(points)